from pathlib import Path
import os
import json

from vocabmaster import utils


def get_config_filepath():
    """
    Gets the configuration file path.

    Returns:
        Path: The path to the application's configuration file.
    """
    app_data_dir = utils.setup_dir()
    config_filepath = app_data_dir / "config.json"
    return config_filepath


def read_config():
    """
    Reads the configuration file.

    Returns:
        dict: The configuration data as a dictionary, or None if the file doesn't exist.
    """
    config_filepath = get_config_filepath()
    if not config_filepath.exists():
        return None
    with open(config_filepath, "r") as file:
        config = json.load(file)
    return config


def write_config(config):
    """
    Writes the configuration data to the configuration file.

    Args:
        config (dict): The configuration data as a dictionary.
    """
    config_filepath = get_config_filepath()
    with open(config_filepath, "w") as file:
        json.dump(config, file, indent=4)


def set_default_language_pair(language_to_learn, mother_tongue):
    """
    Sets the default language pair in the configuration file.

    Args:
        language_to_learn (str): The language the user wants to learn.
        mother_tongue (str): The user's mother tongue.
    """
    config = read_config() or {}
    config["default"] = {
        "language_to_learn": language_to_learn,
        "mother_tongue": mother_tongue,
    }
    write_config(config)


def set_language_pair(language_to_learn, mother_tongue):
    """
    Sets the language pairs in the configuration file.

    Args:
        language_to_learn (str): The language the user wants to learn.
        mother_tongue (str): The user's mother tongue.
    """
    config = read_config() or {"language_pairs": []}
    new_pair = {
        "language_to_learn": language_to_learn,
        "mother_tongue": mother_tongue,
    }
    config["language_pairs"].append(new_pair)
    write_config(config)


def get_default_language_pair():
    """
    Gets the default language pair from the configuration file.

    Returns:
        dict: The default language pair as a dictionary, or None if not found.
    """
    config = read_config()
    if config is None or "default" not in config:
        return None
    return config["default"]


def get_language_pair(language_pair):
    """
    Gets the language pair based on the input option string or the default language pair.

    If the 'pair' argument is not empty, the function will extract the mother tongue and language
    to learn from the input string. If the 'pair' argument is empty, the default language pair
    from the configuration file will be used.

    Args:
        language_pair (str): A string containing the language pair separated by a colon, e.g. "english:french",
            where 'english' is the language to learn and 'french' is the mother tongue. If empty, the default
            language pair from the configuration file will be used.

    Returns:
        tuple: A tuple containing the language to learn and the mother tongue as strings.
    """
    if language_pair:
        try:
            language_to_learn, mother_tongue = language_pair.split(":")
        except ValueError:
            raise ValueError("Invalid language pair.")
    else:
        default_pair = get_default_language_pair()
        if default_pair is None:
            raise ValueError(
                "No default language pair found. Please set a default language pair"
                " using 'vocabmaster config default'.\nSee `vocabmaster --help` for"
                " more information."
            )
        language_to_learn = default_pair["language_to_learn"]
        mother_tongue = default_pair["mother_tongue"]

    return language_to_learn, mother_tongue


def get_all_language_pairs():
    """
    Gets all language pairs from the configuration file.

    Returns:
        list: A list of language pairs as dictionaries.
              The keys are 'language_to_learn' and 'mother_tongue'.
    """
    config = read_config()
    if config is None or "language_pairs" not in config:
        return None
    return config["language_pairs"]



import os
import openai
import tiktoken
import json
import time
from vocabmaster import csv_handler


def format_prompt(language_to_learn, mother_tongue, words_to_translate):
    words_to_translate = "\n".join(words_to_translate)
    prompt = [
        {
            "role": "system",
            "content": """
            You are an expert at building vocabulary lists in a CSV file.
            You do NOT say anything else but the content of the CSV file.""",
        },
        {
            "role": "user",
            "content": f"""
            Translate the following {language_to_learn} words into {mother_tongue}
            and provide a CSV file with each row consisting of the {language_to_learn} word,
            its {mother_tongue} translations (if there are multiple translations possible,
            list them in the same column), and an example sentence in {language_to_learn}.

            Always give ONLY ONE example! The example HAS TO BE in {language_to_learn}!
            Separate each column with commas.
            For the translation column, ALWAYS give at least two or three possible translations!
            
            When you start a new row, you HAVE TO add a newline character.
            Below is the list of words to translate.
            The format should look like this:
            word,'translation1, translation2, translation3','example'.
            ---
            {words_to_translate}""",
        },
    ]
    return prompt


def chatgpt_request(
    prompt,
    model="gpt-3.5-turbo-0301",
    max_tokens=3600,
    n=1,
    temperature=0.5,
    stop=None,
    stream=False,
):
    start_time = time.time()
    openai.api_key = os.getenv("OPENAI_API_KEY")

    # Make the API request
    response = openai.ChatCompletion.create(
        messages=prompt,
        model=model,
        max_tokens=max_tokens,
        n=n,
        temperature=temperature,
        stop=stop,
        stream=stream,
    )

    if stream:
        # Create variables to collect the stream of chunks
        collected_chunks = []
        collected_messages = []

        # Iterate through the stream of events
        for chunk in response:
            collected_chunks.append(chunk)  # save the event response
            chunk_message = chunk["choices"][0]["delta"]  # extract the message
            collected_messages.append(chunk_message)  # save the message
            print(chunk_message.get("content", ""), end="")  # stream the message
        print()
        response = collected_chunks

        # Save the time delay and text received
        response_time = time.time() - start_time
        generated_text = "".join([m.get("content", "") for m in collected_messages])

    else:
        # Extract and save the generated response
        generated_text = response["choices"][0]["message"]["content"]

        # Save the time delay
        response_time = time.time() - start_time

    return (
        generated_text,
        response_time,
        response,
    )


def num_tokens_from_string(string, model="gpt-3.5-turbo-0301"):
    """Returns the number of tokens in a text string."""
    encoding = tiktoken.encoding_for_model(model)
    num_tokens = len(encoding.encode(string))
    return num_tokens


def num_tokens_from_messages(messages, model="gpt-3.5-turbo-0301"):
    """Returns the number of tokens used by a list of messages."""
    try:
        encoding = tiktoken.encoding_for_model(model)
    except KeyError:
        print("Warning: model not found. Using cl100k_base encoding.")
        encoding = tiktoken.get_encoding("cl100k_base")
    if model == "gpt-3.5-turbo":
        print(
            "Warning: gpt-3.5-turbo may change over time. Returning num tokens assuming"
            " gpt-3.5-turbo-0301."
        )
        return num_tokens_from_messages(messages, model="gpt-3.5-turbo-0301")
    elif model == "gpt-4":
        print(
            "Warning: gpt-4 may change over time. Returning num tokens assuming"
            " gpt-4-0314."
        )
        return num_tokens_from_messages(messages, model="gpt-4-0314")
    elif model == "gpt-3.5-turbo-0301":
        tokens_per_message = (
            4  # every message follows <|start|>{role/name}\n{content}<|end|>\n
        )
        tokens_per_name = -1  # if there's a name, the role is omitted
    elif model == "gpt-4-0314":
        tokens_per_message = 3
        tokens_per_name = 1
    else:
        raise NotImplementedError(
            f"""num_tokens_from_messages() is not implemented for model {model}. See https://github.com/openai/openai-python/blob/main/chatml.md for information on how messages are converted to tokens."""
        )
    num_tokens = 0
    for message in messages:
        num_tokens += tokens_per_message
        for key, value in message.items():
            num_tokens += len(encoding.encode(value))
            if key == "name":
                num_tokens += tokens_per_name
    num_tokens += 3  # every reply is primed with <|start|>assistant<|message|>
    return num_tokens


def estimated_cost(num_tokens, price_per_1k_tokens):
    """Returns the estimated cost of a number of tokens."""
    return f"{num_tokens / 1000 * price_per_1k_tokens:.6f}"


def estimate_prompt_cost(message):
    """Returns the estimated cost of a prompt."""
    num_tokens = num_tokens_from_messages(message)

    prices = {
        "gpt-3.5-turbo-0301": 0.002,
        "gpt-4-0314": 0.03,
        "gpt-3.5-turbo": 0.002,
        "gpt-4-8k": 0.03,
        "gpt-4-32k": 0.06,
    }

    return {model: estimated_cost(num_tokens, price) for model, price in prices.items()}



import csv
from csv import DictReader, DictWriter
import platform
from pathlib import Path
from vocabmaster import gpt_integration
from vocabmaster import utils


def word_exists(word, translations_filepath):
    """
    Checks if the word is already present in the `translations_filepath`.

    Args:
        word (str): The word to check for its presence in the file.
        translations_filepath (str): The path to the file containing the list of words.

    Returns:
        bool: True if the word is found in the file, False otherwise.
    """
    with open(translations_filepath, encoding="UTF-8") as file:
        dict_reader = DictReader(file, fieldnames=["word", "translation", "example"])
        for row in dict_reader:
            if word == row["word"]:
                return True
        return False


def append_word(word, translations_filepath):
    """
    Appends the word to the translations file with empty translation and example fields.

    Args:
        word (str): The word to be appended to the file.
        translations_filepath (str): The path to the file containing the list of words.
    """
    with open(translations_filepath, "a", encoding="UTF-8") as file:
        dict_writer = DictWriter(file, fieldnames=["word", "translation", "example"])
        dict_writer.writerow({"word": word, "translation": "", "example": ""})


def get_words_to_translate(translations_filepath):
    """
    Reads a CSV file containing words, translations, and examples, and returns a list of words that need translations.

    The input CSV file should have the following columns: 'word', 'translation', and 'example'.
    If a row is missing either the 'translation' or 'example' column, the 'word' from that row will be added to the list.

    Args:
        translations_filepath (str): The path to the input CSV file containing words, translations, and examples.

    Returns:
        list: A list of words that need translations.
    """
    words_to_translate = []

    with open(translations_filepath, encoding="UTF-8") as translations_file:
        dict_reader = DictReader(translations_file)
        fieldnames = ["word", "translation", "example"]

        for row in dict_reader:
            # If a row is missing a translation or example, add the word to the list of words to translate
            if not row["translation"] or not row["example"]:
                words_to_translate.append(row["word"])

    if not words_to_translate:
        raise Exception(
            "All the words in the vocabulary list already have translations and examples"
        )
    else:
        return words_to_translate


def generate_translations_and_examples(
    language_to_learn, mother_tongue, translations_filepath
):
    """
    Generates translations and examples for a list of words using the GPT model.

    This function calls `get_words_to_translate` to obtain a list of words that need translations,
    using the provided `translations_filepath`. It then formats the prompt using
    `gpt_integration.format_prompt` and sends a request to the GPT model. The generated text from
    the GPT model is returned.

    Args:
        language_to_learn (str): The language to learn.
        mother_tongue (str): The user's mother tongue.
        translations_filepath (str): The path to the input CSV file containing words,
                                       translations, and examples.

    Returns:
        str: The generated text containing translations and examples.
    """
    # Get the list of words that need translations and generate the GPT model prompt
    words_to_translate = get_words_to_translate(translations_filepath)
    prompt = gpt_integration.format_prompt(
        language_to_learn, mother_tongue, words_to_translate
    )

    # Send a request to the GPT model and extract the generated text
    gpt_response = gpt_integration.chatgpt_request(
        prompt=prompt, stream=True, temperature=0.4
    )
    generated_text = gpt_response[0]

    # Create a backup of the GPT response
    backup_dir = utils.get_backup_dir(language_to_learn, mother_tongue)
    utils.backup_content(backup_dir, gpt_response)

    return generated_text


def convert_text_to_dict(generated_text):
    """
    Clean and convert the given text into a dictionary.

    The text should be in the format:
    "'word1',"translation1","example1"\n'word2',"translation2","example2"'

    Args:
        generated_text (str): The text to be cleaned and converted.

    Returns:
        dict: A dictionary with words as keys and a dictionary of translations and examples as values.
    """
    # Clean input text and split it into lines
    cleaned_text = generated_text.replace("\n\n", "")
    lines = cleaned_text.strip().split("\n")

    # Create a dictionary of words with translations and examples
    result = {}
    for line in lines:
        word, translation_and_example = line.split(",", 1)
        translation, example = translation_and_example.rsplit(",", 1)
        result[word.replace("'", "")] = {
            "translation": translation.replace("'", ""),
            "example": example.replace("'", ""),
        }
    return result


def add_translations_and_examples_to_file(translations_filepath, pair):
    """
    Updates the translations file with new translations and examples.

    This function reads a CSV file with words that need translation and their existing translations and examples.
    It generates new translations and examples using the `generate_translations_and_examples` function and updates
    the CSV file with the new translations and examples.

    Args:
        translations_filepath (str): The path to the CSV file containing the translations and examples.
    Returns:
        pair (str): The language pair in the format: 'language_to_learn:mother_tongue'.

        None
    """
    # Generate new translations and examples, then convert the results to a dictionary
    language_to_learn, mother_tongue = utils.get_language_pair_from_option(pair)

    new_entries = convert_text_to_dict(
        generate_translations_and_examples(
            language_to_learn, mother_tongue, translations_filepath
        )
    )

    # Read the current entries from the input file and store them in a dictionary
    with open(translations_filepath, "r", encoding="UTF-8") as input_file:
        translations_reader = DictReader(input_file)
        current_entries = {row["word"]: row for row in translations_reader}

    # Write the updated translations and examples to the output file
    with open(translations_filepath, "w", encoding="UTF-8") as output_file:
        fieldnames = ["word", "translation", "example"]
        writer = DictWriter(output_file, fieldnames=fieldnames)
        writer.writeheader()

        # Iterate through the current entries and update them with the new translations and examples
        for word, current_entry in current_entries.items():
            if word in new_entries and not current_entry["translation"]:
                current_entry["translation"] = new_entries[word]["translation"]
                current_entry["example"] = new_entries[word]["example"]

            # Write the updated entry to the output file
            writer.writerow(current_entry)

    # Create a backup of the translations file
    backup_dir = utils.get_backup_dir(language_to_learn, mother_tongue)
    utils.backup_file(backup_dir, translations_filepath)


def generate_anki_output_file(translations_filepath, anki_output_file):
    """
    Converts a translations file to a CSV file formatted for Anki import.

    This function reads a translations file with words, their translations, and examples, and creates a new CSV file
    formatted as an Anki deck. The resulting file can be imported into Anki to create flashcards with the word on the
    front and the translation and example on the back.

    Args:
        translations_filepath (str): The path to the CSV file containing the translations and examples.
        anki_output_file (str): The path to the output CSV file formatted for Anki import.

    Returns:
        None
    """
    with open(translations_filepath, encoding="UTF-8") as translations_file, open(
        anki_output_file, "w", encoding="UTF-8"
    ) as anki_file:
        translations_dict_reader = DictReader(
            translations_file, fieldnames=["word", "translation", "example"]
        )
        next(translations_dict_reader)

        anki_dict_writer = DictWriter(
            anki_file, fieldnames=["front", "back"], quoting=csv.QUOTE_MINIMAL
        )

        for translations in translations_dict_reader:
            if not translations["translation"] or not translations["example"]:
                continue
            else:
                translations["translation"] = translations["translation"].strip('"')

                # Create a card with the word on the front and the translation and example on the back
                card = {
                    "front": translations["word"],
                    "back": f"{translations['translation']}<br><br><details><summary>example</summary><i>&quot;{translations['example']}&quot;</i></details>",
                }

                # Write the card to the Anki output file
                anki_dict_writer.writerow(card)


def add_fieldnames_to_csv_file(translations_filepath, fieldnames):
    """
    Adds fieldnames to a CSV file if it's missing.

    Args:
        translations_filepath (str): The path to the CSV file.
        fieldnames (list): A list of strings containing the column names.
    """
    with open(translations_filepath, "r+", encoding="UTF-8") as file:
        # Check if the fieldnames is already present in the first row of the content
        for line in file:
            if line.startswith(",".join(fieldnames)):
                return

        file.seek(0, 0)  # Move the file pointer to the beginning of the file
        content = file.read()
        file.seek(0, 0)

        writer = csv.writer(file)
        writer.writerow(fieldnames)  # Write the fieldnames to the first row
        file.write(content)  # Write the original content after the fieldnames


def vocabulary_list_is_empty(translations_filepath):
    """
    Checks if the vocabulary list is empty.

    Args:
        translations_filepath (str): The path to the CSV file containing the translations and examples.

    Returns:
        bool: True if the vocabulary list is empty, False otherwise.
    """
    with open(translations_filepath, encoding="UTF-8") as file:
        csv_reader = csv.reader(file)
        next(csv_reader)  # Skip the fieldnames
        if len(list(csv_reader)) == 0:
            return True



import os
import subprocess
import platform
import shutil
from datetime import datetime
from pathlib import Path
from vocabmaster import config_handler


def setup_dir():
    """
    Creates the application data directory if it doesn't exist and returns its path.
    The directory location is determined based on the global app_name variable and the user's operating system.

    Returns:
        pathlib.Path: The path to the application data directory.
    """
    match platform.system():
        case "Windows":
            app_data_dir = Path.home() / "AppData" / "Roaming" / app_name
        case "Linux" | "Darwin":
            app_data_dir = Path.home() / ".local" / "share" / app_name
        case _:
            print("We couldn't identify your OS.")
            while True:
                try:
                    app_data_dir = Path(
                        input("Please, tell us where you want your files to be installed")
                    )
                    app_data_dir.mkdir(exist_ok=True, parents=True)
                except Exception as e:
                    print(e)
                else:
                    break
    app_data_dir.mkdir(exist_ok=True, parents=True)
    return app_data_dir


def setup_files(app_data_dir, language_to_learn, mother_tongue):
    """
    Creates the necessary file paths in the data directory if they don't exist.

    Args:
    app_data_dir (pathlib.Path): The directory where the application data files should be created.
    language_to_learn (str): The language the user wants to learn.
    mother_tongue (str): The user's mother tongue.

    Returns:
    tuple: A tuple containing the paths of the created files (vocab_list.csv, anki_deck.csv).
    """
    file_paths = (
        app_data_dir / f"vocab_list_{language_to_learn}-{mother_tongue}.csv",
        app_data_dir / f"anki_deck_{language_to_learn}-{mother_tongue}.csv",
    )
    for file in file_paths:
        file.touch()
    return file_paths


def setup_backup_dir(app_data_dir, language_to_learn, mother_tongue):
    """
    Creates the backup directory if it doesn't exist and returns its path.
        The backup directory is created inside the global app_data_dir directory.
        If both mother_tongue and language_to_learn are provided, a subdirectory is created for the specific language pair.

    Args:
        app_data_dir (pathlib.Path): The global application data directory.
        language_to_learn (str): The language the user wants to learn.
        mother_tongue (str): The user's mother tongue.

    Returns:
        pathlib.Path: The path to the backup directory (or the language-specific subdirectory).
    """
    backup_dir = get_backup_dir(language_to_learn, mother_tongue)
    backup_lang = backup_dir / ".backup" / f"{language_to_learn}-{mother_tongue}"
    backup_lang.mkdir(exist_ok=True, parents=True)
    return backup_lang


def get_backup_dir(language_to_learn, mother_tongue):
    """
    Returns the path to the backup directory (or the language-specific subdirectory).

    Args:
        language_to_learn (str): The language the user wants to learn.
        mother_tongue (str): The user's mother tongue.

    Returns:
        pathlib.Path: The path to the backup directory (or the language-specific subdirectory).
    """
    backup_dir = app_data_dir / ".backup"
    if language_to_learn and mother_tongue:
        backup_dir = backup_dir / f"{language_to_learn}-{mother_tongue}"
    return backup_dir


def backup_content(backup_dir, content):
    """
    Creates a timestamped backup file for the provided content in the backup_dir directory.
    A maximum of 15 backup files will be saved, after which the oldest backup file will be overwritten.

    Args:
        backup_dir (pathlib.Path): The backup directory path.
        content (str): The content to be written to the backup file.

    Returns:
        None
    """
    iso_timestamp = generate_iso_timestamp()

    # Create a new backup file with the generated timestamp
    backup_file = backup_dir / f"gpt_request_{iso_timestamp}.bak"
    with backup_file.open("w", encoding="UTF-8") as file:
        file.write(str(content))

    # If there are more than 15 backup files in the directory, delete the oldest backup file
    backup_files = sorted(
        list(backup_dir.glob("gpt_request" + "_*.bak")),
        key=lambda p: p.stat().st_mtime,
    )
    if len(backup_files) > 15:
        oldest_backup_file = backup_files[0]
        oldest_backup_file.unlink(missing_ok=True)


def backup_file(backup_dir, filepath):
    """
    Creates a timestamped backup file for the provided filepath in the backup_dir directory.
    A maximum of 10 backup files will be saved, after which the oldest backup file will be overwritten.

    Args:
        backup_dir (pathlib.Path): The backup directory path.
        filepath (pathlib.Path): The path to the file to be backed up.

    Returns:
        None
    """
    iso_timestamp = generate_iso_timestamp()

    # Create a new backup file with the generated timestamp
    backup_voc_list = backup_dir / f"{filepath.stem}_{iso_timestamp}.bak"
    shutil.copy(filepath, backup_voc_list)

    # If there are more than 10 backup files in the directory, delete the oldest backup file
    backup_files = sorted(
        list(backup_dir.glob(f"{filepath.stem}_*.bak")), key=lambda p: p.stat().st_mtime
    )
    if len(backup_files) > 10:
        oldest_backup_file = backup_files[0]
        oldest_backup_file.unlink()


def generate_iso_timestamp():
    """
    Generates an ISO 8601 formatted timestamp with colons replaced by underscores.

    Returns:
        str: The generated ISO 8601 formatted timestamp with colons replaced by underscores.
    """
    now = datetime.now()
    return now.isoformat().replace(":", "_")


def get_language_pair_from_option(pair):
    """
    Gets the language pair based on the input option string or the default language pair.

    If the 'pair' argument is not empty, the function will extract the mother tongue and language
    to learn from the input string. If the 'pair' argument is empty, the default language pair
    from the configuration file will be used.

    Args:
        pair (str): A string containing the language pair separated by a colon, e.g. "english:french",
            where 'english' is the language to learn and 'french' is the mother tongue. If empty, the default
            language pair from the configuration file will be used.

    Returns:
        tuple: A tuple containing the language to learn and the mother tongue as strings.
    """
    if pair:
        language_to_learn, mother_tongue = pair.split(":")
    else:
        default_pair = config_handler.get_default_language_pair()
        language_to_learn = default_pair["language_to_learn"]
        mother_tongue = default_pair["mother_tongue"]

    return language_to_learn, mother_tongue


def openai_api_key_exists():
    """
    Checks if an OpenAI API key is set on the system.

    Returns:
        bool: True if the OpenAI API key is set, False otherwise.
    """
    return bool(os.environ.get("OPENAI_API_KEY"))


app_name = "vocabmaster"
app_data_dir = setup_dir()
BLUE = "\x1b[94m"
BOLD = "\x1b[1m"
GREEN = "\x1b[92m"
ORANGE = "\x1b[93m"
RED = "\x1b[91m"
RESET = "\x1b[0m"



from setuptools import setup, find_packages


with open("requirements.txt", "r", encoding="utf-8") as file:
    requirements = [line.strip() for line in file]

setup(
    name="vocabmaster",
    version="0.1.0",
    packages=find_packages(),
    install_requires=requirements,
    entry_points={
        "console_scripts": [
            "vocabmaster = vocabmaster.cli:vocabmaster",
        ]
    },
)



import os
import platform
import click
from vocabmaster import csv_handler
from vocabmaster import config_handler
from vocabmaster import gpt_integration
from .utils import *


@click.group()
def vocabmaster():
    """
    VocabMaster is a command-line tool to help you learn vocabulary.

    It uses ChatGPT to generate translations and examples for your words,
    and creates an Anki deck for you to import.

    Start by setting up a new language pair:
    'vocabmaster setup'

    Add words to your vocabulary list:
    'vocabmaster add to have'

    Generate an Anki deck from your vocabulary list:
    'vocabmaster translate'

    You can find help for each command by running:
    'vocabmaster <command> --help'

    For more information, please visit https://github.com/sderev/vocabmaster.
    """
    pass


@vocabmaster.command()
@click.option(
    "--pair",
    type=str,
    help=(
        "This overrides the default language pair. Specify in the format"
        " 'language_to_learn:mother_tongue'. For example: 'english:french'."
    ),
    required=False,
)
@click.argument("word", type=str, nargs=-1)
def add(pair, word):
    """
    Add a word to the vocabulary list, if not already present.

    WORD: The word or phrase to be added to the vocabulary list.

    Examples: 'good', 'to be', 'a cat'
    """
    try:
        language_to_learn, mother_tongue = config_handler.get_language_pair(pair)
    except Exception as error:
        click.echo(f"{RED}Error:{RESET} {error}")
        return

    translations_filepath, anki_filepath = setup_files(
        setup_dir(), language_to_learn, mother_tongue
    )

    if not word:
        click.echo()
        click.echo("Please provide a word to add.")
        click.echo(f"Run '{BOLD}vocabmaster add --help{RESET}' for more information.")
        return

    word = " ".join(word)
    if csv_handler.word_exists(word, translations_filepath):
        click.echo("The word is already in the list 📒")
    else:
        csv_handler.append_word(word, translations_filepath)
        click.echo("The word has been appended to the list 📝✅")


@vocabmaster.command()
@click.option(
    "--pair",
    type=str,
    help=(
        "This overrides the default language pair. Specify in the format"
        " 'language_to_learn:mother_tongue'. For example: 'english:french'."
    ),
    required=False,
)
@click.option(
    "--count",
    is_flag=True,
    help="Show the number of words remaining to be translated in the vocabulary list.",
    required=False,
)
def translate(pair, count):
    """
    Translate, Add examples, and Generate an Anki deck.

    This command reads your vocabulary list, fetches translations and examples,
    and creates an Anki-ready file for import.

    The generated Anki deck will be saved in the same folder as your vocabulary list.
    """
    try:
        language_to_learn, mother_tongue = config_handler.get_language_pair(pair)
    except Exception as error:
        click.echo(f"{RED}Error:{RESET} {error}")
        return

    translations_filepath, anki_filepath = setup_files(
        setup_dir(), language_to_learn, mother_tongue
    )

    # Add the fieldnames to the CSV file if it's missing
    csv_handler.add_fieldnames_to_csv_file(
        translations_filepath, ["word", "translation", "example"]
    )

    # Check if the vocabulary list is empty
    if csv_handler.vocabulary_list_is_empty(translations_filepath):
        click.echo(
            f"{RED}Your vocabulary list is empty.{RESET} Please add some words first."
        )
        click.echo(f"Run '{BOLD}vocabmaster add --help{RESET}' for more information.")
        return

    # Show untranslated words count if `--count` is used, then exit.
    if count:
        try:
            number_words = len(csv_handler.get_words_to_translate(translations_filepath))
            click.echo(f"Number of words to translate: {BLUE}{number_words}{RESET}")
        except Exception as error:
            click.echo(f"{GREEN}Status:{RESET} {error}")
        return

    # Check for OpenAI API key
    if not openai_api_key_exists():
        openai_api_key_explain()
        return

    # Add translations and examples to the CSV file
    click.echo("Adding translations and examples to the file... 🔎📝")
    click.echo(f"{BLUE}This may take a while...{RESET}")
    click.echo()

    try:
        csv_handler.add_translations_and_examples_to_file(translations_filepath, pair)
        click.echo()
    except Exception as error:
        if (
            str(error)
            == "All the words in the vocabulary list already have translations and"
            " examples"
        ):
            click.echo(f"{BLUE}Actually...{RESET}")
            click.echo(f"{GREEN}No action needed:{RESET} {error} 🤓")
            click.echo(
                "If you only want to generate the Anki deck, you can run"
                f" '{BOLD}vocabmaster anki{RESET}'."
            )
        else:
            click.echo(f"{RED}Status:{RESET} {error}")
        return
    click.echo(
        f"{BLUE}The translations and examples have been added to the vocabulary"
        f" list{RESET} 💡✅"
    )

    # Generate the Anki deck
    generate_anki_deck(translations_filepath, anki_filepath)


def generate_anki_deck(translations_filepath, anki_filepath):
    """
    Generates an Anki deck file from a translations file and saves it to the specified path.

    This function takes the input translations file and processes it using the csv_handler's
    generate_anki_output_file function. It then provides feedback to the user about the
    process completion and the location of the generated Anki deck.

    Args:
        translations_filepath (pathlib.Path): Path to the input translations file (CSV format).
        anki_filepath (pathlib.Path): Path to save the generated Anki deck file.

    Returns:
        None
    """
    click.echo()
    click.echo("Generating the Anki deck... 📜")
    click.echo()
    csv_handler.generate_anki_output_file(translations_filepath, anki_filepath)
    click.echo("The Anki deck has been generated 🤓✅")
    click.echo()
    click.echo(f"{GREEN}You can now import the deck into Anki{RESET} 📚")

    click.echo(f"{BOLD}The deck is located at:{RESET}")
    click.echo(f"{anki_filepath}")
    click.echo()


@vocabmaster.command()
def anki():
    """
    Generate an Anki deck from your vocabulary list.

    The Anki deck will be saved in the same folder as your vocabulary list.
    """
    language_to_learn, mother_tongue = config_handler.get_default_language_pair()
    translations_filepath, anki_filepath = setup_files(
        setup_dir(), language_to_learn, mother_tongue
    )
    generate_anki_deck(translations_filepath, anki_filepath)


@vocabmaster.command()
def setup():
    """
    Set up a new language pair.

    This command creates the necessary folders and files
    for the specified mother tongue and language to learn.
    You can set up as many language pairs as you want.

    Example usage:
    vocabmaster setup
    """
    language_to_learn = click.prompt("Please enter the language you want to learn")
    mother_tongue = click.prompt("Please enter your mother tongue")

    click.echo()
    click.echo(
        "Setting up VocabMaster for learning"
        f" {BOLD}{language_to_learn.capitalize()}{RESET}, and"
        f" {BOLD}{mother_tongue.capitalize()}{RESET} is your mother tongue."
    )

    if click.confirm("Do you want to proceed?"):
        # Create the necessary folders and files
        app_data_dir = setup_dir()
        language_to_learn = language_to_learn.casefold()
        mother_tongue = mother_tongue.casefold()

        translations_filepath, anki_filepath = setup_files(
            app_data_dir, language_to_learn, mother_tongue
        )

        backup_lang = setup_backup_dir(app_data_dir, language_to_learn, mother_tongue)
        config_handler.set_language_pair(language_to_learn, mother_tongue)

        click.echo()
        click.echo(f"Translations file: {translations_filepath}")
        click.echo(f"Anki deck file: {anki_filepath}")
        click.echo(f"Backup directory: {backup_lang}")
        click.echo()
        click.echo(
            f"VocabMaster setup for {language_to_learn} to {mother_tongue} complete 🤓✅"
        )
        click.echo()
    else:
        click.echo(f"{RED}Setup canceled{RESET}")

    # Set the default language pair
    if config_handler.get_default_language_pair() is None:
        config_handler.set_default_language_pair(language_to_learn, mother_tongue)
        click.echo(
            f"This language pair ({language_to_learn}:{mother_tongue}) has been set as"
            " the default ✅"
        )

    # Ask the user if they want to set the language pair as the new default
    else:
        print_default_language_pair()

        if click.confirm("Do you want to set this language pair as the default?"):
            if click.confirm(
                f"{RED}Are you sure?{RESET} This will overwrite the current default 🚨"
            ):
                config_handler.set_default_language_pair(language_to_learn, mother_tongue)
            click.echo()
            click.echo("This language pair has been set as the default ✅")
            click.echo(f"{BLUE}The new default language pair is:{RESET}")

            # Get the new default language pair by reinitalizing the variables to avoid confusion
            default_language_to_learn = config_handler.get_default_language_pair()[
                "language_to_learn"
            ]
            default_mother_tongue = config_handler.get_default_language_pair()[
                "mother_tongue"
            ]
            click.echo(
                f"{BOLD}Language to"
                f" learn:{RESET} {default_language_to_learn.capitalize()}"
            )
            click.echo(
                f"{BOLD}Mother tongue:{RESET} {default_mother_tongue.capitalize()}"
            )
            click.echo()

        else:
            click.echo("This language pair has not been set as the default ❌")
            click.echo()
            click.echo("The current default language pair is:")
            default_language_to_learn = config_handler.get_default_language_pair()[
                "language_to_learn"
            ]
            default_mother_tongue = config_handler.get_default_language_pair()[
                "mother_tongue"
            ]
            click.echo(
                f"{BOLD}{default_language_to_learn}:{default_mother_tongue}{RESET}"
            )


@vocabmaster.command()
def default():
    """
    Show the current default language pair.
    """
    print_default_language_pair()
    click.echo()
    click.echo(
        f"{BLUE}You can change the default language pair at any time by running:{RESET}"
    )
    click.echo(f"{BOLD}vocabmaster config default{RESET}")


@vocabmaster.group()
def config():
    """
    Manage the configuration of VocabMaster.

    This command allows you to set the default language pair,
    and the OpenAI API key.

    Example usage:
    'vocabmaster config default'

    Example usage:
    'vocabmaster config key'
    """
    # the directory where the translations and Anki decks are stored,
    #Example usage:
    #'vocabmaster config dir'
    pass


@config.command("default")
def config_default_language_pair():
    """
    Set the default language pair.

    This language pair will be used by default
    when you run the 'vocabmaster' command
    without specifying a language pair with '--pair'.
    """
    print_default_language_pair()

    print_all_language_pairs()
    choice = click.prompt(
        "Type the language pair or its number to set it as the new default",
        type=str,
    )

    # Check if the user entered a correct number
    if choice.isdigit():
        idx = int(choice) - 1
        if 0 <= idx < len(config_handler.get_all_language_pairs()):
            # Set the language pair as the default
            language_to_learn = config_handler.get_all_language_pairs()[idx][
                "language_to_learn"
            ]
            mother_tongue = config_handler.get_all_language_pairs()[idx]["mother_tongue"]
            config_handler.set_default_language_pair(language_to_learn, mother_tongue)
            click.echo(
                f"{BOLD}{language_to_learn}:{mother_tongue}{RESET} {GREEN} has been set"
                f" as the default language pair{RESET} ✅"
            )
        else:
            # The user entered a number that is out of range
            click.echo(f"{RED}Invalid choice{RESET}")
            click.echo(
                "Please enter a number between 1 and"
                f" {len(config_handler.get_all_language_pairs())}"
            )
    else:
        # Check if the language pair exists
        try:
            if config_handler.get_language_pair(choice) is not None:
                # Set the language pair as the default
                language_to_learn, mother_tongue = config_handler.get_language_pair(
                    choice
                )
        # The user entered an invalid language pair
        except ValueError as error:
            click.echo(f"{RED}{error}{RESET}")
            click.echo(f"The format is {BOLD}language_to_learn:mother_tongue{RESET}")
            return

        # Set the language pair as the default
        config_handler.set_default_language_pair(language_to_learn, mother_tongue)
        click.echo(
            f"{BOLD}{language_to_learn}:{mother_tongue}{RESET} {GREEN} has been set as"
            f" the default language pair{RESET} ✅"
        )


#@config.command("dir")
#def config_dir():
#    """
#    Set the directory where the list and the Anki deck are stored.
#    """
#    custom_app_data_dir = click.prompt("Enter the path for the new directory", type=str)
#    setup_dir(custom_app_data_dir)


@config.command("key")
def config_key():
    """
    Set the OpenAI API key.
    """
    if openai_api_key_exists():
        click.echo(f"{GREEN}OpenAI API key found!{RESET}")
        click.echo()
        click.echo(
            f"You can use '{BOLD}vocabmaster translate{RESET}' to generate translations."
        )
        click.echo()
        click.echo(
            "If you only want to generate your Anki deck, you can use"
            f" '{BOLD}vocabmaster anki{RESET}'."
        )
    if not openai_api_key_exists():
        openai_api_key_explain()


def openai_api_key_explain():
    """
    Explain how to set up the OpenAI API key.
    """
    click.echo(f"{RED}You need to set up an OpenAI API key.{RESET}")
    click.echo()
    click.echo(
        "You can generate API keys in the OpenAI web interface. See"
        " https://platform.openai.com/account/api-keys for details."
    )
    click.echo()
    if platform.system() == "Windows":
        click.echo(f"Then, you can set it up by running `setx OPENAI_API_KEY your_key`")
    else:
        click.echo(f"Then, you can set it up by running `export OPENAI_API_KEY=YOUR_KEY`")
    return


@vocabmaster.command()
def show():
    """
    Show all the language pairs that have been set up.
    """
    print_all_language_pairs()
    click.echo(f"{BLUE}You can change the default at any time by running:{RESET}")
    click.echo(f"{BOLD}vocabmaster config default{RESET}")


@vocabmaster.command()
def tokens():
    """
    Estimate the cost of the next translation.

    This command estimates the cost of the next translation,
    based on the number of tokens in the prompt.

    Note that this is only the estimation of the cost of the next prompt,
    not the total cost of the translation.
    The total cost (prompt + translation) cannot exceed $0.008192 per request, though.
    """
    language_to_learn = config_handler.get_default_language_pair()["language_to_learn"]
    mother_tongue = config_handler.get_default_language_pair()["mother_tongue"]
    translations_filepath, anki_file = setup_files(
        setup_dir(), language_to_learn, mother_tongue
    )

    if csv_handler.vocabulary_list_is_empty(translations_filepath):
        click.echo(f"{RED}The list is empty!{RESET}")
        click.echo("Please add words to the list before running this command.")
        return

    try:
        words_to_translate = csv_handler.get_words_to_translate(translations_filepath)
    except Exception as error:
        click.echo(f"{BLUE}Status:{RESET} {error}")
        click.echo("Therefore, the cost of the next prompt cannot be estimated.")
    else:
        prompt = gpt_integration.format_prompt(
            language_to_learn, mother_tongue, words_to_translate
        )
        estimated_cost = gpt_integration.estimate_prompt_cost(prompt)["gpt-3.5-turbo"]
        click.echo(
            f"The estimated cost of the next prompt is {BLUE}${estimated_cost}{RESET}."
        )


def print_default_language_pair():
    """
    Print the current default language pair.
    """
    click.echo()
    click.echo(f"{BLUE}The current default language pair is:{RESET}")
    default_language_to_learn = config_handler.get_default_language_pair()[
        "language_to_learn"
    ]
    default_mother_tongue = config_handler.get_default_language_pair()["mother_tongue"]
    click.echo(
        f"{BOLD}{ORANGE}{default_language_to_learn}:{default_mother_tongue}{RESET}"
    )


def print_all_language_pairs():
    """
    Print all the language pairs that have been set up.
    """
    click.echo()
    click.echo(f"{BLUE}The following language pairs have been set up:{RESET}")
    language_pairs = config_handler.get_all_language_pairs()
    for idx, language_pair in enumerate(language_pairs, start=1):
        click.echo(
            f"{idx}."
            f" {language_pair['language_to_learn']}:{language_pair['mother_tongue']}"
        )
    click.echo()

